"""
This setup.py file sets up our package to be installable on any computer,
so that folks can `import rainbowconnection` from within any directory.
Thanks to this file, you can...

...tell python to look for code in the current directory (which you
can continue to edit), by typing *one* of the following commands:

`pip install -e .`
or
`python setup.py develop`

...move a copy of this code to your site-packages directory, where python will
be able to find it (but you won't be able to keep editing it), by typing *one*
of the following commands:

`pip install .`
or
`python setup.py install`

...upload the entire package to the Python Package Index, so that other folks
will be able to install your package via the simple `pip install rainbow-connection`, by
running the following command:

`python setup.py release`

The template for this setup.py came was pieced together with help from
barentsen, christinahedges, timothydmorton, and dfm. Check them out on github
for more neat tricks!

[`python-packaging`](https://python-packaging.readthedocs.io/) is useful too!
"""

# import our basic setup ingredients
from setuptools import setup, find_packages
import os, sys

# running `python setup.py release` from the command line will post to PyPI
if "release" in sys.argv[-1]:
    os.system("python setup.py sdist")
    # uncomment the next line to test out on test.pypi.com/project/tess-zap
    # os.system("twine upload --repository-url https://test.pypi.org/legacy/ dist/*")
    os.system("twine upload dist/*")
    os.system("rm -rf dist/exoatlas*")
    sys.exit()

# a little kludge to get the version number from __version__
exec(open("exoatlas/version.py").read())

# run the setup function
setup(
    # the name folks can use to search for this with pip
    name="exoatlas",
    # this package will only be installed if the current version doesn't exist
    version=__version__,
    # what's a short description of the package?
    description="Tools for working with transiting exoplanet populations.",
    # what's a more detailed description?
    long_description="For detailed usage, please read the documentation at https://zkbt.github.io/exoatlas/",
    # who's the main author?
    author="Zach Berta-Thompson",
    # what's the main author's email?
    author_email="zach.bertathompson@colorado.edu",
    # what's the URL for the repository?
    url="https://github.com/zkbt/exoatlas",
    # this figures out what subdirectories to include
    packages=find_packages(),
    # are there directories of data that should be accessible when installed?
    include_package_data=True,
    # where are those data directories?
    package_data={},
    # any scripts will be copied into your $PATH, to run from the command line
    scripts=[],
    # some descriptions about this package (for searchability?)
    classifiers=[
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    # what other packages are required. these must be pip-installable
    install_requires=[
        "numpy>=1.13",
        "matplotlib>=2.0",
        "astropy>=3.2.3",
        "astroquery>=0.3.9",
        "astroplan>=0.10",
        "rainbow-connection>=0.0.1",
        "PyYAML",
        "tqdm",
        "pytz",
    ],
    extras_require={
        "develop": [
            "pytest",
            "black",
            "black[jupyter]",
            "jupyter",
            "ipython",
            "mkdocs",
            "mkdocs-material",
            "mkdocstrings",
            "mkdocstrings-python",
            "pytkdocs[numpy-style]",
            "mkdocs-jupyter",
            "mkdocs-exclude",
            "twine",
            "pre-commit",
        ]
    },
    # what version of Python is required?
    python_requires=">=3.9",
    # (I think just leave this set to False)
    zip_safe=False,
    # under what license is this code released?
    license="MIT",
)



from ..imports import *
from .block import Block
from .observatory import Observatory
from .transit import Transit
from matplotlib import animation


# a list of colors to give transits (will repeat)
colors = [
    "salmon",
    "firebrick",
    "hotpink",
    "mediumvioletred",
    "coral",
    "orangered",
    "darkorange",
    "plum",
    "fuchsia",
    "darkviolet",
    "purple",
    "indigo",
    "slateblue",
    "limegreen",
    "seagreen",
    "darkgreen",
    "olive",
    "teal",
    "steelblue",
    "royalblue",
    "sandybrown",
    "chocolate",
    "saddlebrown",
    "maroon",
]


class Plan:
    """a Plan object to keep track of potentially observable transits"""

    def __init__(
        self,
        pop=None,
        start=None,
        finish=None,
        observatory="SBO",
        maxairmass=2.5,
        maxsun=-12.0,
        buffer=1.0,
        allow_partial=True,
        directory="upcoming-transits/",
        name="default",
        file_format="png",
    ):
        """initialize, setting the observatory, block, and population"""

        # set the maximum airmass and the maximum sun altitude
        self.maxairmass = maxairmass
        self.maxsun = maxsun
        self.allow_partial = allow_partial
        self.buffer = buffer

        self.file_format = file_format
        # set up a directory to store results in
        self.directory = directory
        mkdir(self.directory)

        # set up our observatory
        self.observatory = Observatory(observatory, plan=self)

        # set up the time range this should cover
        self.block = Block(start=start, finish=finish, plan=self)

        # load the population of planets
        self.population = pop

        # tidy up that population
        self.tidy_population()

        # print out a summary of this plan
        self.describe_inputs()

    def describe_inputs(self):
        """print text to terminal describing this observing plan"""

        print("Observatory is {}".format(self.observatory))
        print("Time range spans {} to {}".format(self.block.start, self.block.finish))
        print("Population contains {} objects".format(len(self.population)))

    def tidy_population(self):
        """load a population of exoplanets, defaulting to the list
        of confirmed transiting planets from the NASA Exoplanet Archive"""

        # remove planets that are impossible to see from this latitude
        zenith = np.abs(self.population.dec() - self.observatory.latitude)
        possible = zenith < 60.0 * u.deg
        print(
            "{}/{} targets are visible from {} latitude".format(
                np.sum(possible), len(possible), self.observatory.latitude
            )
        )

        # trim the population
        self.population = self.population[possible]

        print(
            "the trimmed population contains {} objects".format(
                len(self.population.standard)
            )
        )

        # KLUDGE
        bad_duration = np.isfinite(self.population.transit_duration()) == False

        # print('These planets have bad durations!')
        # print(self.population[bad_duration].name)

        self.population = self.population[bad_duration == False]
        # add some colors to the population
        N = len(self.population)
        c = np.array(colors)[np.arange(N) % len(colors)]
        self.population.standard["observing_color"] = c

    def find_planets_transiting_on_night(self, midnight):
        """identify all the transits for all the planets"""

        now = midnight
        pop = self.population

        # pull out the period and epoch of all the planets
        P = pop.period().to("day")
        T0 = pop.transit_midpoint().to("day")
        coords = SkyCoord(ra=pop.ra(), dec=pop.dec())

        # find the closest exact transit epochs
        pop.standard["closest_epoch_number"] = np.round((now - T0) / P)
        pop.standard["closest_epoch_day"] = pop.closest_epoch_number * P + T0

        # ask which planets have transits happening tonight/today
        dt = now - pop.closest_epoch_day
        is_happening_tonight = np.abs(dt) < 0.5 * u.day
        sum(is_happening_tonight), len(is_happening_tonight)

        # pull out only the times where a transit is happening within 0.5 days
        times_happening = Time(pop.closest_epoch_day[is_happening_tonight], format="jd")
        buffer = self.buffer * pop.transit_duration[is_happening_tonight]
        coords_happening = coords[is_happening_tonight]

        # figure out if the targets are up
        min_alt = (
            90 * u.deg - np.arccos(1 / self.maxairmass) * 180 / np.pi * u.deg
        )  # 30*u.deg
        up_happening = (
            self.observatory.altaz(coords_happening, times_happening).alt > min_alt
        )
        if self.allow_partial == False:
            up_happening *= (
                self.observatory.altaz(coords_happening, times_happening + buffer).alt
                > min_alt
            )
            up_happening *= (
                self.observatory.altaz(coords_happening, times_happening - buffer).alt
                > min_alt
            )

        # figure out if the sun is down
        max_sun = self.maxsun * u.deg  # -12*u.deg
        dark_happening = self.observatory.sun(times_happening).alt < max_sun
        dark_happening *= self.observatory.sun(times_happening + buffer).alt < max_sun
        dark_happening *= self.observatory.sun(times_happening - buffer).alt < max_sun

        # populate a mask of which transits we should both thinking about
        should_consider = True
        should_consider *= is_happening_tonight
        should_consider[is_happening_tonight] *= up_happening
        should_consider[is_happening_tonight] *= dark_happening

        planets_tonight = pop[should_consider]
        return planets_tonight

    def plot_transits_on_night(self, midnight):
        # plot black bars to mask out when the Sun is up
        self.observatory.plot_empty_night(midnight)

        planets = self.find_planets_transiting_on_night(midnight)
        for p in planets:
            t = Transit(p, p.closest_epoch_number, observatory=self.observatory)
            t.plot(color=p.observing_color[0])

        plt.ylim(2.3, 0.9)

    def plot_all_transits(self):
        """
        Plot all observable transits that fall within this night.
        """

        for midnight in tqdm(self.block.midnights):
            try:
                self.plot_transits_on_night(midnight)
                date = Time(
                    midnight - self.observatory.standardzone - 0.5 * u.day, format="jd"
                ).iso[:10]
                plt.title(f"Transits Observable from {self.observatory.name} on {date}")

                filename = os.path.join(self.directory, f"{date}.{self.file_format}")
                plt.savefig(filename)
                plt.clf()
            except ValueError:
                print(f"Something went wrong on {midnight}!")

    def print_transits(self, filename="upcoming_transits.txt"):
        """print all the observable transits"""
        with open(filename, "w") as f:
            for p in self.planets:
                p._speak(p.name)
                for t in p.transits:
                    s = t.details()
                    f.write(s + "\n")
                # (t.posttransit - t.pretransit).value*24))
                # self.observatory.sun(t.pretransit).alt.deg[0],
                # self.observatory.sun(t.posttransit).alt.deg[0]))

    def movieTransits(self, filename=None, fps=10, bitrate=10000):
        """
        Make a movie of the whole block,
        showing all the transits for your population.
        """

        print("plotting transits")
        self.find_transits()
        plt.ioff()

        # set up the plotting window (one night at a time)
        self.figure = plt.figure(
            "upcoming transits", figsize=(15, 6), dpi=200, constrained_layout=True
        )
        self.gs = plt.matplotlib.gridspec.GridSpec(
            2, 1, hspace=0, wspace=0, height_ratios=[0.3, 1.0], bottom=0.35
        )
        self.ax = {}

        # an ax to include the names of the planets
        self.ax["transits"] = plt.subplot(self.gs[0])
        plt.setp(self.ax["transits"].get_xticklabels(), visible=False)
        plt.setp(self.ax["transits"].get_yticklabels(), visible=False)
        self.ax["transits"].plot_date([], [])
        plt.setp(self.ax["transits"].get_xaxis(), visible=False)
        plt.setp(self.ax["transits"].get_yaxis(), visible=False)
        self.ax["transits"].set_title(
            "Observability from {0}".format(self.observatory.name, self.block.name)
        )

        # an ax to include the airmass curves
        self.ax["airmass"] = plt.subplot(self.gs[1], sharex=self.ax["transits"])
        self.ax["airmass"].set_ylabel("Airmass")

        # plot all the individual transits for all the planets
        for i in range(len(self.planets)):
            self.planets[i].plotTransits(y=i)

        # nudge the limits on the transit labels
        # self.ax['transits'].set_xlim(self.block.start.plot_date, self.block.finish.plot_date)
        self.ax["transits"].set_ylim(-0.1 * len(self.planets), 1.2 * len(self.planets))

        # plot black bars to mask out when the Sun is up
        for k in self.ax.keys():
            self.observatory.plotSun(jd=self.block.times, ax=self.ax[k])

        # set the formatting on the time label
        fmt = plt.matplotlib.dates.DateFormatter("%Y-%m-%d\n%H:%M UT")
        self.ax["airmass"].xaxis.set_major_formatter(fmt)
        f = plt.gcf()
        f.autofmt_xdate(rotation=90)

        print("making a movie of transits")

        # initialize the animator
        metadata = dict(
            title="Observability from {0}".format(self.observatory.name),
            artist="whatsup",
        )
        # self.writer = animation.FFMpegWriter(fps=fps, metadata=metadata, bitrate=bitrate)
        self.writer = animation.PillowWriter(
            fps=fps, metadata=metadata, bitrate=bitrate
        )
        if filename is None:
            filename = self.directory + "transits-{}planets.mp4".format(
                len(self.planets)
            )
        print("trying to save movie to {0}".format(filename))
        with self.writer.saving(self.figure, filename, self.figure.get_dpi()):
            # loop over exposures
            window = 7 / 24.0

            # approximate kludge; no daylight savings; need to incorporate time zones
            offset = self.observatory.standardzone.to("day").value

            for m in tqdm(self.block.midnights):
                xlim = self.ax["airmass"].get_xlim()
                center = Time(m, format="jd").plot_date + offset
                self.ax["airmass"].set_xlim((center - window), (center + window))

                nightof = Time(m - 0.5, format="jd").iso.split()[0]
                self.ax["transits"].set_title(
                    "Observability from {} on {}".format(self.observatory.name, nightof)
                )
                self.writer.grab_frame()



from ...imports import *
from .exoplanet_downloaders import *
from IPython.display import display

from astropy.table import join

kws = {
    "ps": dict(color="orchid"),
    "default": dict(color="black"),
    "pscp": dict(color="royalblue"),
}


class NASAExoplanetArchiveComparison:

    # define some variables to pay attention to in plots
    some_key_parameters = dict(
        stellar_basics=["st_mass", "st_rad", "st_dens", "st_lum", "st_teff"],
        stellar_fancy=["st_logg", "st_vsin", "st_met", "st_rotp", "st_age"],
        planet_basics=["pl_rade", "pl_bmasse", "pl_dens", "pl_orbper", "pl_orbsmax"],
        planet_fancy=[
            "pl_insol",
            "pl_eqt",
            "pl_imppar",
            "pl_ratror",
            "pl_ratdor",
            "pl_trueobliq",
            "pl_rvamp",
            "pl_orblper",
            "pl_orbtper",
            "pl_orbincl",
        ],
        planet_transit=[
            "pl_trandur",
            "pl_tranmid",
            "pl_trandep",
        ],
        stellar_positions=[
            "ra",
            "dec",
            "sy_dist",
            "sy_plx",
            "sy_pmra",
            "sy_pmdec",
            "st_radv",
        ],
    )

    def __init__(
        self, plot_directory="nasa-exoplanet-archive-table-comparisons", **downloadkw
    ):

        # empty dictionary of
        self.tables = {}

        # download both tables from NASA Exoplanet Archive
        self.tables["pscp"] = ExoplanetArchiveDownloader("pscomppars").get_table(
            **downloadkw
        )
        self.tables["ps"] = ExoplanetArchiveDownloader("ps").get_table(**downloadkw)

        # trim down to just default solutions
        is_default = self.tables["ps"]["default_flag"] == 1
        self.tables["default"] = self.tables["ps"][is_default]

        # create joined tables to be able to compare
        self.joined = {}
        for k in ["ps", "default"]:
            self.joined[f"pscp+{k}"] = join(
                self.tables["pscp"],
                self.tables[k],
                keys="pl_name",
                table_names=["pscp", k],
            )

        # create a directory to store plots
        self.plot_directory = plot_directory
        try:
            os.mkdir(plot_directory)
        except FileExistsError:
            pass

    def plot_parameter_comparison(self, k="pl_rade"):
        """
        Visualize how a parameter compares across the complete Planetary Systems,
        the default Planetary Systems, and Planetary Systems Composite Parameters.

        Parameters
        ----------
        k : str
            Which parameter to visualize.
        """
        plotkw = dict(marker=".", linewidth=0, alpha=0.5)

        fi, ax = plt.subplots(
            2,
            2,
            sharex="col",
            sharey="row",
            figsize=(8, 5),
            dpi=600,
            constrained_layout=True,
            gridspec_kw=dict(width_ratios=[1, 0.5]),
        )

        # plot one vs the other
        plt.sca(ax[0, 0])
        plt.title(k)
        for t in ["ps", "default"]:
            plt.loglog(
                self.joined[f"pscp+{t}"][f"{k}_pscp"],
                self.joined[f"pscp+{t}"][f"{k}_{t}"],
                label=t,
                **plotkw,
                **kws[t],
            )
        plt.ylabel(f'"{k}" from Planetary Systems')
        plt.legend(frameon=False, loc="upper left")

        # plot ratio, as points and as sideways histogram
        for t in ["ps", "default"]:
            ratio = (
                self.joined[f"pscp+{t}"][f"{k}_{t}"]
                / self.joined[f"pscp+{t}"][f"{k}_pscp"]
            )
            plt.sca(ax[1, 1])
            plt.hist(
                ratio, orientation="horizontal", bins=np.logspace(-3, 3, 300), **kws[t]
            )
            plt.xscale("log")

            plt.sca(ax[1, 0])
            plt.loglog(
                self.joined[f"pscp+{t}"][f"{k}_pscp"],
                ratio,
                label="all",
                **plotkw,
                **kws[t],
            )
        plt.xlabel(f'"{k}" from Planetary Systems Composite Parameters')
        plt.ylabel(f"ps/pscp")
        plt.ylim(0.5e-3, 2e3)
        plt.axhline(0.1, color="gray", linestyle="--")

        # display a text summary
        plt.sca(ax[0, 1])
        plt.axis("off")
        s = f"{k}:\n\n"
        for t in ["ps", "default", "pscp"]:
            this = self.tables[t][k]
            try:
                N = np.sum(this.mask == False)
            except AttributeError:
                N = np.sum(np.isfinite(this))
            s += f"{t} has data for {N/len(this):>5.1%} ({N}/{len(this)})\n"
        s += "\n"
        for t in ["ps", "default"]:
            this_joined = self.joined[f"pscp+{t}"]
            is_same = this_joined[f"{k}_{t}"] == this_joined[f"{k}_pscp"]
            N = np.sum(is_same)
            s += f"{t} == pscp  for {N/len(this_joined):>5.1%} ({N}/{len(this_joined)})\n"
        plt.text(0, 1, s, transform=ax[0, 1].transAxes, va="top", ha="left", fontsize=8)
        return s

    def do_completeness_check(self):
        """
        Check how complete each table is for some key parameters.
        """
        for group in self.some_key_parameters:
            # print(group)
            for k in self.some_key_parameters[group]:
                s = self.plot_parameter_comparison(k)
                # print(s)
                plt.savefig(
                    os.path.join(self.plot_directory, f"check-completeness-{k}.png")
                )
            # print

    def do_units_check(self, egregious_threshold=0.05):
        """
        Check how values compare when provided in Earth or Jupiter units.
        """
        # loop over radius and mass
        for p, q in zip(["rad", "bmass"], ["R", "M"]):

            plt.figure(figsize=(8, 5), constrained_layout=True)
            plt.title(f"'{p}' in different units")

            # loop over tables
            for k in ["ps", "default", "pscp"]:
                from_earth = self.tables[k][f"pl_{p}e"] * u.Unit(f"{q}earth")
                from_jupiter = self.tables[k][f"pl_{p}j"] * u.Unit(f"{q}jup")
                ratio = (from_earth / from_jupiter).decompose()
                plt.semilogx(
                    from_earth,
                    ratio,
                    marker=".",
                    linewidth=0,
                    alpha=0.25,
                    label=k,
                    **kws[k],
                )
                plt.ylabel(f"(pl_{p}e/pl_{p}j)" + r"$\times (M_{Jupiter}/M_{Earth})$")
                plt.ylim(0.75, 1.25)

                if k == "pscp":
                    is_egregious = np.abs(ratio - 1) > egregious_threshold
                    bad_planets = self.tables[k][is_egregious]["pl_name"]
                    print(
                        f"""
                    The parameter '{p}' is off by more than {egregious_threshold:%} for the {len(bad_planets)} planets 
                    {bad_planets}
                    simply because of weird rounding and/or Earth-Jupiter unit conversions.
                    """
                    )
                    print(list(self.tables[k][is_egregious]["pl_name"]))
            plt.legend(frameon=False)
            plt.savefig(os.path.join(self.plot_directory, f"check-units-{p}.png"))

            plt.show()

    def do_mass_radius_check(self):
        """
        Check that theoretical masses/radii get removed.
        """
        kw = dict(
            marker=".",
            linewidth=0,
            alpha=0.25,
        )
        plt.figure(figsize=(8, 5), constrained_layout=True)
        for k in ["ps", "default", "pscp"]:
            ok_mass_error = (self.tables[k]["pl_bmasseerr1"].mask == False) * (
                self.tables[k]["pl_bmasseerr2"].mask == False
            )
            ok_radius_error = (self.tables[k]["pl_radeerr1"].mask == False) * (
                self.tables[k]["pl_radeerr2"].mask == False
            )
            t = self.tables[k][ok_mass_error * ok_radius_error]
            plt.loglog(t["pl_bmasse"], t["pl_rade"], label=k, **kws[k], **kw)

        t = self.tables[k][ok_mass_error * ok_radius_error == False]
        plt.loglog(
            t["pl_bmasse"],
            t["pl_rade"],
            color="red",
            label="no mass|radius errors",
            **kw,
        )

        plt.axis("scaled")
        plt.legend(frameon=False)
        plt.xlabel(r"Planet Mass ($M_\oplus$)")
        plt.ylabel(r"Planet Radius ($R_\oplus$)")
        plt.savefig(os.path.join(self.plot_directory, f"check-mass-radius.png"))

    def do_luminosity_check(self):
        kw = dict(
            marker=".",
            linewidth=0,
            alpha=0.25,
        )
        fi, ax = plt.subplots(
            2, 1, figsize=(8, 5), constrained_layout=True, sharex=True
        )
        for k in ["ps", "default", "pscp"]:
            t = self.tables[k]  # [ok_mass_error * ok_radius_error]
            ok_lum = (t["st_lumerr1"].mask == False) * (t["st_lumerr2"].mask == False)
            ok_rad = (t["st_raderr1"].mask == False) * (t["st_raderr2"].mask == False)
            ok_teff = (t["st_tefferr1"].mask == False) * (
                t["st_tefferr2"].mask == False
            )
            ok = ok_lum * ok_rad * ok_lum
            T = t["st_teff"] * u.K
            L = 10 ** t["st_lum"] * u.Lsun
            R = t["st_rad"] * u.Rsun
            plt.sca(ax[0])
            x = (4 * np.pi * R**2 * con.sigma_sb * T**4).to(u.Lsun)
            y = L.to(u.Lsun)
            plt.loglog(x[ok], y[ok], label=k, **kws[k], **kw)
            plt.sca(ax[1])
            plt.semilogx(x[ok], (y / x)[ok], label=k, **kws[k], **kw)
            plt.ylim(0, 2)

            bad = ok == False
            plt.sca(ax[0])
            plt.loglog(
                x[bad],
                y[bad],
                label=f"{k} (no lum/teff/radius uncertainty)",
                **kws[k],
                **kw,
            )
            plt.plot([1e-6, 1e4], [1e-6, 1e4], linestyle="--", color="gray")
            plt.ylabel("L")
            plt.legend(frameon=False, bbox_to_anchor=(1, 1), loc="upper left")

            plt.sca(ax[1])
            plt.semilogx(x[bad], (y / x)[bad], **kws[k], **kw)
            plt.axhline(1, color="gray", linestyle="--")
            plt.ylabel(r"L / [$4\pi R^2 \sigma T_{eff}^4$]")

        plt.sca(ax[1])
        plt.xlabel(r"$4\pi R^2 \sigma T_{eff}^4$")
        plt.savefig(
            os.path.join(self.plot_directory, f"check-luminosity-radius-teff.png")
        )



"""
Tools for working with the habitable zone definitions
in Kopparapu et al. (2013).
"""

coefficients = {}
coefficients["recent-venus"] = [1.7753, 1.4316e-4, 2.9875e-9, -7.5702e-12, -1.1635e-15]
coefficients["runaway-greenhouse"] = [
    1.0512,
    1.3242e-4,
    1.5418e-8,
    -7.9895e-12,
    -1.8328e-15,
]
coefficients["moist-greenhouse"] = [
    1.0140,
    8.1774e-5,
    1.7063e-9,
    -4.3241e-12,
    -6.6462e-16,
]
coefficients["maximum-greenhouse"] = [
    0.348,
    5.8924e-5,
    1.6558e-9,
    -3.0045e-12,
    -5.2983e-16,
]
coefficients["early-mars"] = [0.3179, 5.4513e-5, 1.5313e-9, -2.7786e-12, -4.8997e-16]


def make_hz(which="runaway-greenhouse"):
    """
    Create a function f(Teff) describing one of
    the habitable zone boundaries presented in
    Table 3 of Kopparapu et al. (2013).
    """

    # pull out the coefficients
    try:
        S_o, a, b, c, d = coefficients[which]
    except KeyError:
        error_message = f"""
        Alas! There seems to be no way to determine
        coefficients for your choice of `{which}`.
        The available options are:
        """
        for k in which:
            error_message += k + "\n"
        raise ValueError(error_message)

    # define the function
    def f(Teff):
        f"""
        The `{which}` HZ boundary from Kopparapu et al. (2013),
        as a function of stellar effective temperature.
        """

        T = Teff - 5780
        return S_o + a * T + b * T**2 + c * T**3 + d * T**4

    # return that function definition
    return f
